# To import required modules:
import numpy as np
from matplotlib.colors import LogNorm # for log color scales
from scipy.special import erf # error function, used in computing CDF of normal distribution





# Useful fundamental constants:

AU = 1.496*10.**13. # AU in cm
Msun = 1.989*10.**30. # Solar mass in kg
Rsun = 6.957*10.**10. # Solar radius in cm
Mearth = 5.972*10.**24. # Earth mass in kg
Rearth = 6.371*10.**8. # Earth radius in cm
Mjup = 1.898*10.**27. # Jupiter mass in kg
Rjup = 6.991*10.**9. # Jupiter radius in cm





# Miscellaneous definitions:

res_ratios, res_width = [2.0, 1.5, 4/3., 5/4.], 0.05 # NOTE: in the model, the near-resonant planets have period ratios between X and (1+w)*X where X = [2/1, 3/2, 4/3, 5/4] and w = 0.05





# Miscellaneous functions:

def a_from_P(P, Mstar):
    # Convert period (days) to semi-major axis (AU) assuming mass of planet m << Mstar (Msun)
    """
    Compute the semi-major axis using Kepler's third law.

    Note
    ----
    Assumes that the planet mass is negligible compared to the stellar mass.

    Parameters
    ----------
    P : float or array[floats]
        The orbital period (days).
    Mstar : float or array[floats]
        The stellar mass (solar masses).

    Returns
    -------
    a : float or array[floats]
        The semi-major axis (AU).
    """
    a = (P/365.25)**(2./3.)*(Mstar/1.0)**(1./3.)
    return a

def P_from_a(a, Mstar):
    """
    Compute the orbital period using Kepler's third law.

    Note
    ----
    Assumes that the planet mass is negligible compared to the stellar mass.

    Parameters
    ----------
    a : float or array[floats]
        The semi-major axis (AU).
    Mstar : float or array[floats]
        The stellar mass (solar masses).

    Returns
    -------
    P : float or array[floats]
        The orbital period (days).
    """
    P = 365.25*(a**(3./2.))*(Mstar/1.0)**(-1./2.)
    return P

def M_from_R_rho(R, rho=5.51):
    """
    Compute the planet mass from the radius and constant mean density.

    Parameters
    ----------
    R : float or array[floats]
        The planet radius (Earth radii).
    rho : float or array[floats], default=5.51
        The planet density (g/cm^3). Default value is that of Earth.

    Returns
    -------
    M : float or array[floats]
        The planet mass (Earth masses).
    """
    M_in_g = rho * (4.*np.pi/3.)*(R*Rearth)**3.
    return M_in_g/(Mearth*1000.)

def rho_from_M_R(M, R):
    """
    Compute the planet mean density (total mass divided by volume).

    Parameters
    ----------
    M : float or array[floats]
        The planet mass (Earth masses).
    R : float or array[floats]
        The planet radius (Earth radii).

    Returns
    -------
    rho : float or array[floats]
        The planet mean density (g/cm^3).
    """
    rho = (M*Mearth*1000.) / ((4.*np.pi/3.)*(R*Rearth)**3.)
    return rho

def tdur_circ(P, Mstar, Rstar):
    """
    Compute the transit duration assuming a circular orbit with zero impact parameter.

    Parameters
    ----------
    P : float or array[floats]
        The orbital period (days).
    Mstar : float or array[floats]
        The stellar mass (solar masses).
    Rstar : float or array[floats]
        The stellar radius (solar radii).

    Returns
    -------
    tdur : float or array[floats]
        The transit duration (hrs).
    """
    tdur = 24.*(Rstar*Rsun*P)/(np.pi*a_from_P(P,Mstar)*AU)
    return tdur

def AMD(mu, a, e, im):
    """
    Compute the AMD (angular momentum deficit) of a planet(s).

    Note
    ----
    Uses units of ``G*M_star = 1``.

    Parameters
    ----------
    mu : float or array[floats]
        The planet/star mass ratio.
    a : float or array[floats]
        The semi-major axis (AU).
    e : float or array[floats]
        The orbital eccentricity.
    im : float or array[floats]
        The inclination relative to the system invariable plane (radians).

    Returns
    -------
    amd_pl : float or array[floats]
        The AMD of the planet(s).
    """
    amd_pl = mu*np.sqrt(a) * (1. - np.sqrt(1. - e**2.)*np.cos(im))
    return amd_pl

def NAMD(m, a, e, im):
    """
    Compute the NAMD (normalized angular momentum deficit) of a planetary system.

    First defined in Chambers 2001, the NAMD of a system is the AMD divided by the angular momentum of the same system with circular and coplanar orbits.

    Parameters
    ----------
    m : float or array[floats]
        The planet masses (Earth masses).
    a : float or array[floats]
        The semi-major axes (AU).
    e : float or array[floats]
        The orbital eccentricities.
    im : float or array[floats]
        The inclinations relative to the system invariable plane (radians).

    Returns
    -------
    normed_AMD : float
        The NAMD of the system (unitless).
    """
    num = np.sum(m*np.sqrt(a) * (1. - np.sqrt(1. - e**2.)*np.cos(im)))
    den = np.sum(m*np.sqrt(a))
    normed_AMD = num/den
    return normed_AMD

def photoevap_boundary_Carrera2018(R, P):
    """
    Evaluate whether a planet is above or below the 'photo-evaporation' valley defined by Eq. 5 in Carrera et al. (2018).

    Parameters
    ----------
    R : float
        The planet radius (Earth radii).
    P : float
        The orbital period (days).

    Returns
    -------
    above_boundary : 1 or 0
        Whether the planet is above (1) or below (0) the boundary.
    """
    Rtrans = 2.6*P**(-0.1467)
    if R >= Rtrans:
        above_boundary = 1
    else:
        above_boundary = 0
    return above_boundary

def incl_mult_power_law_Zhu2018(k, sigma_5=0.8, alpha=-3.5):
    """
    Compute the Rayleigh scale of the mutual inclination distribution for a given planet multiplicity using the power-law relation.

    Note
    ----
    Default values for the power-law parameters ``sigma_5`` and ``alpha`` are set to best-fit values from Zhu et al. (2018).

    Parameters
    ----------
    k : float or array[floats]
        The planet multiplicity.
    sigma_5 : float or array[floats], default=0.8
        The normalization (Rayleigh scale of the mutual inclination distribution at ``k=5``) (degrees).
    alpha : float or array[floats], default=-3.5
        The power-law index.

    Returns
    -------
    sigma_k : float or array[floats]
        The Rayleigh scale (degrees) of the mutual inclination distribution for multiplicity ``k``.
    """
    sigma_k = sigma_5*(k/5.)**alpha
    return sigma_k

def cdf_normal(x, mu=0., std=1.):
    """
    Compute the cumulative distribution function (CDF; i.e. the integral between ``-inf`` and ``x`` of) the normal distribution at ``x`` given a mean and standard deviation.

    Note
    ----
    Can deal with array inputs for ``x``, ``mu``, and ``std`` as long as they are the same shape.

    Parameters
    ----------
    x : float or array[floats]
        The position to evaluate the CDF (between ``-inf`` and ``inf``).
    mu : float or array[floats], default=0.
        The mean of the normal distribution.
    std : float or array[floats], default=1.
        The standard deviation of the normal distribution.

    Returns
    -------
    cdf_x : float or array[floats]
        The CDF at ``x``.
    """
    cdf_x = 0.5*(1. + erf((x - mu)/(std*np.sqrt(2))))
    return cdf_x

def cdf_empirical(xdata, xeval):
    """
    Compute the empirical cumulative distribution function (CDF) at ``xeval`` given a sample.

    Note
    ----
    Is designed to deal with either scalar or array inputs of ``xeval``.

    Parameters
    ----------
    xdata : array
        The sample of data points.
    xeval : float or array[floats]
        The position to evaluate the CDF.

    Returns
    -------
    cdf_at_xeval : float or array[floats]
        The CDF at ``xeval``.
    """
    N = len(xdata)
    xeval = np.asarray(xeval)
    is_xeval_scalar = False
    if xeval.ndim == 0:
        xeval = xeval[None] # turn x_eval into array with 1 element
        is_xeval_scalar = True

    cdf_at_xeval = np.zeros(len(xeval))
    for i,x in enumerate(xeval):
        cdf_at_xeval[i] = np.sum(xdata <= x)/N

    if is_xeval_scalar:
        return np.squeeze(cdf_at_xeval)
    return cdf_at_xeval

def calc_f_near_pratios(sssp_per_sys, pratios=res_ratios, pratio_width=res_width):
    """
    Compute the intrinsic fraction of planets 'near' a period ratio with another planet for any period ratio in the given list of period ratios.

    'Near' is defined as a period ratio between ``pr`` and ``pratio*(1+pratio_width)`` for ``pratio`` in ``pratios``.

    Note
    ----
    Defaults to calculating the fraction of all planets near a mean-motion resonance (MMR), defined by ``res_ratios``.

    Parameters
    ----------
    sssp_per_sys : dict
        The dictionary of summary statistics for a physical catalog of planets.
    pratios : list, default=res_ratios
        The list of period ratios to consider.
    pratio_width : float, default=res_width
        The fractional width to be considered 'near' a period ratio.

    Returns
    -------
    f_mmr : float
        The fraction of planets being 'near' at least one of the period ratios.
    """
    count_mmr = 0
    for p_sys in sssp_per_sys['P_all']:
        p_sys = p_sys[p_sys > 0]
        mmr_sys = np.zeros(len(p_sys))
        pr_sys = p_sys[1:]/p_sys[:-1]
        pr_mmr_sys = np.zeros(len(pr_sys))

        for mmr in res_ratios:
            pr_mmr_sys[(pr_sys >= mmr) & (pr_sys <= mmr*(1.+res_width))] = 1
        for j,res in enumerate(pr_mmr_sys):
            if res == 1:
                mmr_sys[j] = 1
                mmr_sys[j+1] = 1
        count_mmr += np.sum(mmr_sys == 1)
    f_mmr = float(count_mmr)/np.sum(sssp_per_sys['P_all'] > 0)
    return f_mmr

def compute_ratios_adjacent(x, inverse=False, avoid_div_zeros=False):
    """
    Compute an array of the adjacent ratios (e.g. ``x[j+1]/x[j]``) of the terms in the input array ``x``.

    Parameters
    ----------
    x : array[float]
        The input values.
    inverse : bool, default=False
        Whether to take the inverse ratios (i.e. ``x[j]/x[j+1]`` instead of ``x[j+1]/x[j]``).
    avoid_div_zeros : bool, default=False
        Whether to avoid dividing by zeros. If True, will avoid computing ratios where the denominator is zero, and insert either 'inf' (when the numerator is non-zero) or 'nan' (when the numerator is also zero). If False, the computed ratios will still have the same results, but will produce ``RuntimeWarning`` messages.

    Returns
    -------
    ratios_adj : array[float]
        The ratios of adjacent pairs in ``x``.


    Note
    ----
    Will return an empty array if there are fewer than two elements in the input array.
    """
    if len(x) <= 1:
        return np.array([])

    if avoid_div_zeros and any(x==0):
        ratios_adj = np.zeros(len(x)-1)
        numers = x[0:-1] if inverse else x[1:] # the numerators
        denoms = x[1:] if inverse else x[0:-1] # the denominators
        bools_denoms_nonzero = denoms != 0 # booleans indicating the denominators that are non-zero
        bools_denoms_zero = ~bools_denoms_nonzero
        bools_numers_zero = numers == 0 # booleans indicating the numerators that are zero (for finding where we have 0/0)
        ratios_adj[bools_denoms_nonzero] = numers[bools_denoms_nonzero] / denoms[bools_denoms_nonzero]
        ratios_adj[bools_denoms_zero & ~bools_numers_zero] = np.sign(numers[bools_denoms_zero & ~bools_numers_zero])*np.inf # put +/-inf (depending on the sign of the numerator) when only denominator is zero
        ratios_adj[bools_denoms_zero & bools_numers_zero] = np.nan # put nans when both denominator and numerator are zero
        return ratios_adj

    # If get to this point, 'x' has at least two elements and none are zero
    ratios_adj = x[0:-1]/x[1:] if inverse else x[1:]/x[0:-1]
    return ratios_adj

def compute_ratios_all(x, inverse=False, avoid_div_zeros=False):
    """
    Compute an array of all the unique ratios (``x[j]/x[i]`` for all ``j > i``) of the terms in the input array ``x``.

    Parameters
    ----------
    x : array[float]
        The input values.
    inverse : bool, default=False
        Whether to take the inverse ratios (i.e. ``x[j]/x[i]`` for all ``j < i`` instead of for all ``j > i``).
    avoid_div_zeros : bool, default=False
        Whether to avoid dividing by zeros (WARNING: not implemented yet!).

    Returns
    -------
    ratios_all : array[float]
        The ratios of all pairs in ``x``.


    Note
    ----
    Will return an empty array if there are fewer than two elements in the input array.
    """
    if len(x) <= 1:
        return np.array([])

    ratios_all = list(x[0]/x[1:]) if inverse else list(x[1:]/x[0])
    for i in range(len(x)-2):
        ratios_all += list(x[i+1]/x[i+2:]) if inverse else list(x[i+2:]/x[i+1])
    return np.array(ratios_all)

def zeta1(pratios):
    """Compute the zeta statistic for each period ratio in ``pratios`` as defined in Fabrycky et al. (2014)."""
    return 3.*((1./(pratios - 1.)) - np.round(1./(pratios - 1.)))

def split_colors_per_cdpp_bin(stars_cleaned, nbins=10):
    # Compute a histogram of combined differential photometric precision (CDPP) values, and then split each bin by `bp-rp` color into a bluer half (smaller `bp-rp`) and a redder half (larger `bp-rp`).
    """
    Return the indices of the bluer and redder stars that split the stellar sample into two samples based on the histogram of combined differential photometric precision (CDPP) values.

    Note
    ----
    Uses log-uniform bins for the histograms.

    Parameters
    ----------
    stars_cleaned : structured array
        A table of stars, including columns for 4.5hr CDPP (``rrmscdpp04p5``) and Gaia DR2 bp-rp color (``bp_rp``).
    nbins : int, default=10
        The number of bins to use.

    Returns
    -------
    bins : array[floats]
        The bin edges.
    i_blue_per_bin : array[floats]
        The indices corresponding to the rows of bluer stars in each bin.
    i_red_per_bin : array[floats]
        The indices corresponding to the rows of redder stars in each bin.
    """
    cdpp_min, cdpp_max = np.min(stars_cleaned['rrmscdpp04p5']), np.max(stars_cleaned['rrmscdpp04p5'])
    bp_rp_med = np.nanmedian(stars_cleaned['bp_rp'])
    counts, bins = np.histogram(stars_cleaned['rrmscdpp04p5'], bins=np.logspace(np.log10(cdpp_min), np.log10(cdpp_max), nbins+1))
    i_blue_per_bin, i_red_per_bin = [], []
    for i in range(nbins):
        i_bin = np.where((stars_cleaned['rrmscdpp04p5'] >= bins[i]) & (stars_cleaned['rrmscdpp04p5'] < bins[i+1]))[0]
        bp_rp_bin = stars_cleaned['bp_rp'][i_bin]
        bp_rp_bin_med = np.nanmedian(bp_rp_bin)
        i_bin_blue = i_bin[bp_rp_bin <= bp_rp_bin_med]
        i_bin_red = i_bin[bp_rp_bin > bp_rp_bin_med]
        i_blue_per_bin += list(i_bin_blue)
        i_red_per_bin += list(i_bin_red)
    i_blue_per_bin = np.array(i_blue_per_bin)
    i_red_per_bin = np.array(i_red_per_bin)
    return bins, i_blue_per_bin, i_red_per_bin

def linear_fswp_bprp(bprp, bprp_med, fswp_med=0.5, slope=0.):
    """
    Evaluate the fraction of stars with planets (fswp) at a number of bp-rp colors using a linear relation.

    Note
    ----
    The fswp cannot be negative or greater than one.

    Parameters
    ----------
    bprp : array[floats]
        The bp-rp colors at which to evaluate the fswp.
    bprp_med : float
        The median bp-rp color (or some normalization point).
    fswp_med : float, default=0.5
        The fswp at ``bprp_med`` (normalization).
    slope : float, default=0.
        The slope of the linear relation.

    Returns
    -------
    fswp_bprp : array[floats]
        The fswp at each bp-rp color.
    """
    bprp = np.array(bprp)
    fswp_bprp = slope*(bprp - bprp_med) + fswp_med
    fswp_bprp[fswp_bprp < 0] = 0.
    fswp_bprp[fswp_bprp > 1] = 1.
    return fswp_bprp

def linear_alphaP_bprp(bprp, bprp_med, alphaP_med=0.5, slope=0.):
    """
    Evaluate the period power-law index at a number of bp-rp colors using a linear relation.

    Parameters
    ----------
    bprp : array[floats]
        The bp-rp colors at which to evaluate the fswp.
    bprp_med : float
        The median bp-rp color (or some normalization point).
    alphaP_med : float, default=0.5
        The period power-law index at ``bprp_med`` (normalization).
    slope : float, default=0.
        The slope of the linear relation.

    Returns
    -------
    alphaP_bprp : array[floats]
        The period power-law index at each bp-rp color.
    """
    bprp = np.array(bprp)
    alphaP_bprp = slope*(bprp - bprp_med) + alphaP_med
    return alphaP_bprp

def bin_Nmult(Nmult_obs, m_geq=5):
    """
    Bins the higher orders of the input multiplicity distribution.

    Parameters
    ----------
    Nmult_obs : array[ints]
        The multiplicity distribution for consecutive multiplicity orders (i.e. number of systems with 1,2,3,... planets).
    m_geq : int, default=5
        The multiplicity order at and above which to bin together.

    Returns
    -------
    Nmult_obs : array[ints]
        The multiplicity distribution with multiplicity orders greater than or equal to ``m_geq`` binned together. For the default ``m_geq=5``, this is the number of systems with 1,2,3,4,5+ planets.
    """
    Nmult_obs = list(Nmult_obs) + [0]*(m_geq-len(Nmult_obs)) # zero-pad first
    Nmult_obs[m_geq-1] = np.sum(Nmult_obs[m_geq-1:]) # bin everything greater than or equal to m_geq
    return np.array(Nmult_obs[:m_geq])





# Information theory quantities and metrics from Gilbert & Fabrycky 2020 (GF2020):

def Shannon_entropy(p):
    """
    Compute the Shannon entropy.

    Note
    ----
    Assumes base-natural log, although any choice is valid.

    Parameters
    ----------
    p : array[floats]
        An array of probabilities (values between 0 and 1).

    Returns
    -------
    H : float
        The Shannon entropy.
    """
    assert all(p >= 0), 'Negative probabilities!'
    assert all(p <= 1), 'Probabilities greater than 1!'

    H = -np.sum(p*np.log(p))
    return H

def disequilibrium(p):
    """
    Compute the disequilibrium of a system.

    Parameters
    ----------
    p : array[floats]
        An array of probabilities (values between 0 and 1).

    Returns
    -------
    D : float
        The disequilibrium.
    """
    assert all(p >= 0), 'Negative probabilities!'
    assert all(p <= 1), 'Probabilities greater than 1!'

    D = np.sum((p - (1./len(p)))**2.)
    return D

def LMC_complexity(K, p):
    """
    Compute the Lopez-Ruiz, Mancini, and Calbet (1995) complexity, which is the product of Shannon entropy and disequilibrium.

    Parameters
    ----------
    K : float
        The normalization constant.
    p : array[floats]
        An array of probabilities (values between 0 and 1).

    Returns
    -------
    C : float
        The LMC complexity.
    """
    assert K > 0

    H = Shannon_entropy(p)
    D = disequilibrium(p)
    C = K*H*D
    return C

def Pearson_correlation_coefficient(x, y):
    """Compute the Pearson correlation coefficient of two variables."""
    assert len(x) == len(y)

    xmean, ymean = np.mean(x), np.mean(y)
    r_xy = np.sum((x - xmean)*(y - ymean)) / np.sqrt(np.sum((x - xmean)**2.)*np.sum((y - ymean)**2.))
    return r_xy

def Spearman_correlation_coefficient(x, y):
    """
    Compute the Spearman correlation coefficient of two variables.

    This is the Pearson correlation coefficient applied to the ranks of the variables.
    """
    xsort, ysort = np.argsort(x), np.argsort(y)
    xranks, yranks = np.zeros(len(x)), np.zeros(len(y))
    xranks[xsort], yranks[ysort] = np.arange(len(x)), np.arange(len(y))
    rho_S = Pearson_correlation_coefficient(xranks, yranks)
    return rho_S



def radii_star_ratio(r, Rstar):
    """
    Compute the sum of planet/star radius ratios for a system.

    Parameters
    ----------
    r : array[floats]
        The planet radii.
    Rstar : float
        The stellar radius.

    Returns
    -------
    mu : float
        The sum of the planet/star radius ratios.
    """
    assert all(r >= 0), 'Negative planet radii!'
    assert Rstar > 0, 'Negative stellar radii!'

    mu = np.sum(r/Rstar)
    return mu

def partitioning(x):
    """
    Compute the 'partitioning' of a quantity ``x`` for the system.

    For example, if ``x`` is an array of planet masses, this is the 'mass partitioning' quantity.

    Parameters
    ----------
    x : array[floats]
        The partitions of the quantity.

    Returns
    -------
    Q : float
        The partitioning metric for the quantity (between 0 and 1).
    """
    assert all(x >= 0), 'Negative x values!'

    xnorm = x/np.sum(x)
    Q = (len(x)/(len(x)-1.))*disequilibrium(xnorm) # factor of (N/(N-1)) should normalize Q to (0,1)
    return Q

def monotonicity_GF2020(x):
    """
    Compute the 'monotonicity' (a measure of the degree of ordering) of a quantity ``x`` for the system, defined in Gilbert and Fabrycky (2020).

    For example, if ``x`` is an array of planet radii, this is the 'radius monotonicity' quantity.

    Parameters
    ----------
    x : array[floats]
        The array of quantities.

    Returns
    -------
    M : float
        The monotonicity of the quantity.
    """
    rho_S = Spearman_correlation_coefficient(np.arange(len(x)), x)
    Q = partitioning(x)
    M = rho_S*(Q**(1./len(x)))
    return M

def gap_complexity_GF2020(P):
    """
    Compute the 'gap complexity' metric of a planetary system, as defined in Gilbert and Fabrycky (2020).

    Parameters
    ----------
    P : array[floats]
        The array of orbital periods (days). Does not have to be sorted.

    Returns
    -------
    C : float
        The gap complexity of the system.
    """
    assert len(P) >= 3, 'Need at least 3 planets in system to compute.'
    n = len(P)-1

    P = np.sort(P)
    Pmin, Pmax = np.min(P), np.max(P)
    Pratios = P[1:]/P[:-1]

    pnorm = np.log(Pratios)/np.log(Pmax/Pmin) # assuming natural log is what GF2020 used?
    Cmax = Cmax_table_GF2020(n) if n < 10 else Cmax_approx_GF2020(n)
    K = 1./Cmax

    C = LMC_complexity(K, pnorm)
    return C

def Cmax_table_GF2020(n):
    """
    Return the value of 'C_max', based on the number of gaps ``n`` (i.e. the number of planets minus 1) in the system.

    This is a normalization used in the gap complexity metric, with the values computed by Gilbert and Fabrycky (2020).

    Note
    ----
    The table of 'C_max' only goes up to ``n=9``; for greater numbers of gaps, use the approximation function :py:func:`syssimpyplots.general.Cmax_approx_GF2020`.
    """
    Cmax_dict = {2: 0.106, 3: 0.212, 4: 0.291, 5: 0.350, 6: 0.398, 7: 0.437, 8: 0.469, 9: 0.497}
    return Cmax_dict[n]

def Cmax_approx_GF2020(n):
    """Compute an approximation for the value of 'C_max' based on the number of gaps, defined in Gilbert and Fabrycky (2020)."""
    Cmax = 0.262*np.log(0.766*n)
    return Cmax





# Useful functions for general purposes:

class MidPointLogNorm(LogNorm):
    """
    Set the midpoint of a colormap on a log scale, taken from: https://stackoverflow.com/questions/48625475/python-shifted-logarithmic-colorbar-white-color-offset-to-center (in a post by ImportanceOfBeingErnest)
    """
    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        LogNorm.__init__(self,vmin=vmin, vmax=vmax, clip=clip)
        self.midpoint=midpoint
    def __call__(self, value, clip=None):
        result, is_scalar = self.process_value(value)
        x, y = [np.log(self.vmin), np.log(self.midpoint), np.log(self.vmax)], [0, 0.5, 1]
        return np.ma.array(np.interp(np.log(value), x, y), mask=result.mask, copy=False)



# To import required modules:
import numpy as np
import os

import syssimpyplots.general as gen

path_data = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')





# Functions to load and analyze the Kepler observed catalog:

N_Kep = 86760 #86760 (Paper III) #88912 (Paper II) #79935 (Paper I) # number of Kepler targets satisfying our cuts to give our observed catalog # TODO: automate or read this from a file (stellar catalog?) instead of hardcoding

def load_Kepler_planets_cleaned():
    """
    Load a table of the Kepler objects of interest (KOIs) from a CSV file.

    Returns
    -------
    planets_cleaned : structured array
        A table with the properties of the KOIs.


    The table has the following columns:

    - `kepid`: The Kepler ID.
    - `KOI`: The KOI number.
    - `koi_disposition`: The disposition of the KOI.
    - `koi_pdisposition`: (TODO: TBD).
    - `koi_score`: The disposition score (between 0 and 1).
    - `P`: The orbital period (days).
    - `t_D`: The transit duration (hrs).
    - `depth`: The transit depth (ppm).
    - `Rp`: The planet radius (Earth radii).
    - `teff`: The stellar effective temperature (K).
    - `logg`: The log surface gravity of the star.
    - `Rstar`: The stellar radius (solar radii).
    - `Mstar`: The stellar mass (solar masses).

    """
    # q1_q17_dr25_gaia_fgk_HFR2021a_koi_cleaned.csv for Paper II
    # q1_q17_dr25_gaia_berger_fgk_H2020_koi_cleaned.csv for Paper III
    planets_cleaned = np.genfromtxt(os.path.join(path_data, 'q1_q17_dr25_gaia_berger_fgk_H2020_koi_cleaned.csv'), dtype={'names': ('kepid', 'KOI', 'koi_disposition', 'koi_pdisposition', 'koi_score', 'P', 't_D', 'depth', 'Rp', 'teff', 'logg', 'Rstar', 'Mstar'), 'formats': ('i8', 'S9', 'S15', 'S15', 'f8', 'f8', 'f8', 'f8', 'f8', 'f8', 'f8', 'f8', 'f8',)}, delimiter=',') #orbit periods 'P' are in days; transit durations 't_D' are in hrs; transit depths 'depth' are in ppm; planetary radii 'Rp' are in Rearth; stellar radii 'Rstar' are in Rsolar
    planets_cleaned = planets_cleaned[1:]
    return planets_cleaned

def load_Kepler_stars_cleaned():
    """
    Load a table of Kepler target stars from a CSV file.

    Returns
    -------
    stars_cleaned : structured array
        A table with the properties of the Kepler target stars.


    The table has the following columns:

    - `kepid`: The Kepler ID.
    - `mass`: The stellar mass (solar masses).
    - `radius`: The stellar radius (solar radii).
    - `teff`: The stellar effective temperature (K).
    - `bp_rp`: The Gaia DR2 bp-rp color (mag).
    - `lum_val`: The luminosity (solar luminosities).
    - `e_bp_rp_interp`: The extinction in bp-rp color (mag) interpolated from a model/binning.
    - `e_bp_rp_true`: The extinction in bp-rp color (mag) as given in the Gaia DR2 catalog.
    - `rrmscdpp04p5`: The root-mean-square combined differential photometric precision (CDPP) for 4.5 hr durations (ppm).

    """
    # q1_q17_dr25_gaia_fgk_HFR2021a_cleaned.csv for Paper II
    # q1_q17_dr25_gaia_berger_fgk_H2020_cleaned.csv for Paper III
    #stars_cleaned = np.genfromtxt(os.path.join(path_data, 'q1_q17_dr25_gaia_fgk_HFR2021a_cleaned.csv'), dtype={'names': ('kepid', 'mass', 'radius', 'teff', 'bp_rp', 'e_bp_rp_interp', 'e_bp_rp_true', 'rrmscdpp04p5'), 'formats': ('i8', 'f8', 'f8', 'f8', 'f8', 'f8', 'f8', 'f8')}, delimiter=',')
    stars_cleaned = np.genfromtxt(os.path.join(path_data, 'q1_q17_dr25_gaia_berger_fgk_H2020_cleaned.csv'), dtype={'names': ('kepid', 'mass', 'radius', 'teff', 'bp_rp', 'lum_val', 'e_bp_rp_interp', 'e_bp_rp_true', 'rrmscdpp04p5'), 'formats': ('i8', 'f8', 'f8', 'f8', 'f8', 'f8', 'f8', 'f8', 'f8')}, delimiter=',')
    stars_cleaned = stars_cleaned[1:]
    return stars_cleaned

def compute_summary_stats_from_Kepler_catalog(P_min, P_max, radii_min, radii_max, Rstar_min=0., Rstar_max=10., Mstar_min=0., Mstar_max=10., teff_min=0., teff_max=1e4, bp_rp_min=-5., bp_rp_max=5., i_stars_custom=None, compute_ratios=gen.compute_ratios_adjacent):
    """
    Compute detailed summary statistics per system in the Kepler catalog.

    Parameters
    ----------
    P_min : float
        The minimum orbital period (days) to be included.
    P_max : float
        The maximum orbital period (days) to be included.
    radii_min : float
        The minimum planet radius (Earth radii) to be included.
    radii_max : float
        The maximum planet radius (Earth radii) to be included.
    Rstar_min : float, default=0.
        The minimum stellar radius (solar radii) to be included.
    Rstar_max : float, default=10.
        The maximum stellar radius (solar radii) to be included.
    Mstar_min : float, default=0.
        The minimum stellar mass (solar masses) to be included.
    Mstar_max : float, default=10.
        The maximum stellar mass (solar masses) to be included.
    teff_min : float, default=0.
        The minimum stellar effective temperature (K) to be included.
    teff_max : float, default=0.
        The maximum stellar effective temperature (K) to be included.
    bp_rp_min : float, default=-5.
        The minimum Gaia DR2 bp-rp color (mag) to be included.
    bp_rp_max : float, default=5.
        The maximum Gaia DR2 bp-rp color (mag) to be included.
    i_stars_custom : array[int], default=None
        An array of indices for the stars in the Kepler stellar catalog to be included.
    compute_ratios : func, default=compute_ratios_adjacent
        The function to use for computing ratios; can be either :py:func:`syssimpyplots.general.compute_ratios_adjacent` or :py:func:`syssimpyplots.general.compute_ratios_all`.

    Returns
    -------
    ssk_per_sys : dict
        A dictionary containing the planetary and stellar properties for each observed system (2-d and 1-d arrays).
    ssk : dict
        A dictionary containing the planetary and stellar properties of all observed planets (1-d arrays).


    The fields of ``ssk_per_sys`` and ``ssk`` are the same as those returned by :py:func:`syssimpyplots.load_sims.compute_summary_stats_from_cat_obs`.
    """
    planets_cleaned = load_Kepler_planets_cleaned()
    stars_cleaned = load_Kepler_stars_cleaned()

    if i_stars_custom is not None:
        stars_cleaned = stars_cleaned[i_stars_custom]

        i_pl_from_stars = []
        for i,kepid in enumerate(planets_cleaned['kepid']):
            i_stars_cleaned = np.where(stars_cleaned['kepid'] == kepid)[0]
            if len(i_stars_cleaned) == 1:
                i_pl_from_stars.append(i)
        i_pl_from_stars = np.array(i_pl_from_stars)
        planets_cleaned = planets_cleaned[i_pl_from_stars]

    # NOTE: these include duplicates for stars with multiple planets:
    n_pl = len(planets_cleaned)
    Rstar_per_pl = np.zeros(n_pl)
    Mstar_per_pl = np.zeros(n_pl)
    teff_per_pl = np.zeros(n_pl)
    bp_rp_per_pl = np.zeros(n_pl)
    e_bp_rp_per_pl = np.zeros(n_pl)
    cdpp4p5_per_pl = np.zeros(n_pl)
    for i,kepid in enumerate(planets_cleaned['kepid']):
        i_stars_cleaned = np.where(stars_cleaned['kepid'] == kepid)[0][0]
        Rstar_per_pl[i] = stars_cleaned['radius'][i_stars_cleaned]
        Mstar_per_pl[i] = stars_cleaned['mass'][i_stars_cleaned]
        teff_per_pl[i] = stars_cleaned['teff'][i_stars_cleaned]
        bp_rp_per_pl[i] = stars_cleaned['bp_rp'][i_stars_cleaned]
        e_bp_rp_per_pl[i] = stars_cleaned['e_bp_rp_interp'][i_stars_cleaned]
        cdpp4p5_per_pl[i] = stars_cleaned['rrmscdpp04p5'][i_stars_cleaned]

    planets_cleaned['Rstar'] = Rstar_per_pl
    planets_cleaned['Mstar'] = Mstar_per_pl
    planets_cleaned['teff'] = teff_per_pl

    # To make cuts in period, planet radii, and stellar properties:
    indices_keep = np.arange(len(planets_cleaned))[(planets_cleaned['P'] >= P_min) &
                                                   (planets_cleaned['P'] < P_max) &
                                                   (planets_cleaned['Rp'] >= radii_min) &
                                                   (planets_cleaned['Rp'] < radii_max) &
                                                   (Rstar_per_pl >= Rstar_min) &
                                                   (Rstar_per_pl < Rstar_max) &
                                                   (Mstar_per_pl >= Mstar_min) &
                                                   (Mstar_per_pl < Mstar_max) &
                                                   (teff_per_pl >= teff_min) &
                                                   (teff_per_pl < teff_max) &
                                                   (bp_rp_per_pl - e_bp_rp_per_pl >= bp_rp_min) &
                                                   (bp_rp_per_pl - e_bp_rp_per_pl < bp_rp_max)]

    planets_cleaned = planets_cleaned[indices_keep]



    Rstar_cand = Rstar_per_pl[indices_keep] #planets_cleaned['Rstar']
    Mstar_cand = Mstar_per_pl[indices_keep] #planets_cleaned['Mstar']
    teff_cand = teff_per_pl[indices_keep] #planets_cleaned['teff']
    bp_rp_cand = bp_rp_per_pl[indices_keep]
    e_bp_rp_cand = e_bp_rp_per_pl[indices_keep]
    cdpp4p5_cand = cdpp4p5_per_pl[indices_keep]

    # To keep arrays for the stellar properties only counting each star once:
    Rstar_sys_cand = []
    Mstar_sys_cand = []
    teff_sys_cand = []
    bp_rp_sys_cand = []
    e_bp_rp_sys_cand = []
    cdpp4p5_sys_cand = []



    #To compute the arrays of observables:
    KOI_systems = np.array([x[:6] for x in planets_cleaned['KOI']])
    checked_bools = np.zeros(len(planets_cleaned)) #0's denote KOI that were not checked yet; 1's denote already checked KOI

    P_sys_cand = [] # periods per system
    D_sys_cand = [] # depths per system
    tdur_sys_cand = [] # durations per system
    tdur_tcirc_sys_cand = [] # circular normalized durations per system
    radii_sys_cand = [] # planet radii per system
    Rm_sys_cand = [] # period ratios per system
    D_ratio_sys_cand = [] # depth ratios per system
    xi_sys_cand = [] # period-normalized duration ratios per system

    M_cand = [] # planet multiplicities of systems with planets
    Rm_cand = [] # period ratios
    D_ratio_cand = [] # depth ratios
    xi_cand = [] # period-normalized duration ratios
    xi_res_cand = [] # period-normalized duration ratios of planet pairs near resonance
    xi_res32_cand = [] # period-normalized duration ratios of planet pairs near 3:2 resonance
    xi_res21_cand = [] # period-normalized duration ratios of planet pairs near 2:1 resonance
    xi_nonres_cand = [] # period-normalized duration ratios of planet pairs not near resonance
    tdur_cand = planets_cleaned['t_D'] # durations (hrs)

    D_cand = planets_cleaned['depth']/(1e6) # depths (fraction)
    radii_cand = planets_cleaned['Rp'] # planet radii (Rearth)
    D_above_cand = [] # depths of planets above the photoevaporation boundary
    D_below_cand = [] # depths of planets below the photoevaporation boundary
    D_ratio_above_cand = [] # depth ratios of planet pairs above the photoevaporation boundary
    D_ratio_below_cand = [] # depth ratios of planet pairs below the photoevaporation boundary
    D_ratio_across_cand = [] # depth ratios of planet pairs across the photoevaporation boundary

    pad_extra = 100 #####

    for i in range(len(KOI_systems)):
        if checked_bools[i] == 0: #if the KOI has not been checked (included while looking at another planet in the same system)
            system_i = np.where(KOI_systems == KOI_systems[i])[0]
            checked_bools[system_i] = 1
            system_Rstar = planets_cleaned['Rstar'][system_i][0]
            system_Mstar = planets_cleaned['Mstar'][system_i][0]
            system_teff = planets_cleaned['teff'][system_i][0]
            system_bp_rp = bp_rp_cand[system_i][0]
            system_e_bp_rp = e_bp_rp_cand[system_i][0]
            system_cdpp4p5 = cdpp4p5_cand[system_i][0]

            #To get the planet properties in this system:
            system_P = planets_cleaned['P'][system_i] #periods of all the planets in this system
            system_tdur = planets_cleaned['t_D'][system_i] #transit durations of all the planets in this system
            system_D = planets_cleaned['depth'][system_i]/(1e6) #transit depths of all the planets in this system
            system_radii = planets_cleaned['Rp'][system_i] #radii of all the planets in this system
            system_sort_i = np.argsort(system_P) #indices that would sort the periods of the planets in this system
            system_P = system_P[system_sort_i] #periods of all the planets in this system, sorted
            system_tdur = system_tdur[system_sort_i] #transit durations of all the planets in this system, sorted by period
            system_D = system_D[system_sort_i] #transit depths of all the planets in this system, sorted by period
            system_radii = system_radii[system_sort_i] #radii of all the planets in this system, sorted by period

            system_tdur_tcirc = system_tdur/gen.tdur_circ(system_P, system_Mstar, system_Rstar)

            system_Rm = compute_ratios(system_P) #period ratios of all the adjacent planet pairs in this system
            system_D_ratio = compute_ratios(system_D) #transit depth ratios of all the adjacent planet pairs in this system
            system_xi = compute_ratios(system_tdur, inverse=True, avoid_div_zeros=True)*compute_ratios(system_P)**(1./3.) #period-normalized transit duration ratios of all the adjacent planet pairs in this system

            # To append the stellar properties counting the star only once:
            Rstar_sys_cand.append(system_Rstar)
            Mstar_sys_cand.append(system_Mstar)
            teff_sys_cand.append(system_teff)
            bp_rp_sys_cand.append(system_bp_rp)
            e_bp_rp_sys_cand.append(system_e_bp_rp)
            cdpp4p5_sys_cand.append(system_cdpp4p5)

            # To count the total number of planets in this system:
            M_cand.append(len(system_P))

            # To save arrays of planet properties per system:
            P_sys_cand.append(list(system_P) + [-1]*(pad_extra - len(system_P)))
            D_sys_cand.append(list(system_D) + [-1]*(pad_extra - len(system_D)))
            tdur_sys_cand.append(list(system_tdur) + [-1]*(pad_extra - len(system_tdur)))
            tdur_tcirc_sys_cand.append(list(system_tdur_tcirc) + [np.nan]*(pad_extra - len(system_tdur_tcirc)))
            radii_sys_cand.append(list(system_radii) + [-1]*(pad_extra - len(system_radii)))
            Rm_sys_cand.append(list(system_Rm) + [-1]*(pad_extra - len(system_Rm)))
            D_ratio_sys_cand.append(list(system_D_ratio) + [-1]*(pad_extra - len(system_D_ratio)))
            xi_sys_cand.append(list(system_xi) + [-1]*(pad_extra - len(system_xi)))

            # To separate into planet pairs near vs. not in resonance:
            mask_res_system = np.zeros(len(system_Rm), dtype=bool)
            mask_res32_system = np.zeros(len(system_Rm), dtype=bool)
            mask_res21_system = np.zeros(len(system_Rm), dtype=bool)

            for ratio in gen.res_ratios:
                mask_res_system[(system_Rm >= ratio) & (system_Rm <= ratio*(1.+gen.res_width))] = 1

            mask_res32_system[(system_Rm >= 1.5) & (system_Rm <= 1.5*(1.+gen.res_width))] = 1
            mask_res21_system[(system_Rm >= 2.) & (system_Rm <= 2.*(1.+gen.res_width))] = 1
            system_xi_res = system_xi[mask_res_system]
            system_xi_res32 = system_xi[mask_res32_system]
            system_xi_res21 = system_xi[mask_res21_system]
            system_xi_nonres = system_xi[~mask_res_system]
            #if sum(mask_res_system) > 0:
            #print(system_R[mask_res_system], system_xi_res)

            for Rm in system_Rm:
                Rm_cand.append(Rm)
            for D_ratio in system_D_ratio:
                D_ratio_cand.append(D_ratio)
            for xi in system_xi:
                xi_cand.append(xi)
            for xi in system_xi_res:
                xi_res_cand.append(xi)
            for xi in system_xi_res32:
                xi_res32_cand.append(xi)
            for xi in system_xi_res21:
                xi_res21_cand.append(xi)
            for xi in system_xi_nonres:
                xi_nonres_cand.append(xi)

            #To separate the planets in the system as above and below the boundary:
            system_above_bools = np.array([gen.photoevap_boundary_Carrera2018(system_radii[x], system_P[x]) for x in range(len(system_P))])
            #if len(system_above_bools) > 1:
            #print(system_above_bools)

            #To record the transit depths of the planets above and below the boundary:
            for j in range(len(system_D)):
                D_above_cand.append(system_D[j]) if system_above_bools[j] == 1 else D_below_cand.append(system_D[j])

            #To record the transit depth ratios of the planets above, below, and across the boundary:
            if compute_ratios == gen.compute_ratios_adjacent:
                for j in range(len(system_D_ratio)):
                    if system_above_bools[j] + system_above_bools[j+1] == 2: #both planets are above the boundary
                        D_ratio_above_cand.append(system_D_ratio[j])
                    elif system_above_bools[j] + system_above_bools[j+1] == 1: #one planet is above, the other is below the boundary
                        D_ratio_across_cand.append(system_D_ratio[j])
                    elif system_above_bools[j] + system_above_bools[j+1] == 0: #both planets are below the boundary
                        D_ratio_below_cand.append(system_D_ratio[j])
            elif compute_ratios == gen.compute_ratios_all:
                ncum_ratio_j = np.concatenate((np.array([0]),np.cumsum(np.arange(len(system_D))[::-1][:-2])))
                for j in range(len(system_D_ratio)):
                    for k in range(j+1,len(system_D)):
                        i_ratio = ncum_ratio_j[j] + k - 1 - j
                        if system_above_bools[j] + system_above_bools[k] == 2: #both planets are above the boundary
                            D_ratio_above_cand.append(system_D_ratio[i_ratio])
                        elif system_above_bools[j] + system_above_bools[k] == 1: #one planet is above, the other is below the boundary
                            D_ratio_across_cand.append(system_D_ratio[i_ratio])
                        elif system_above_bools[j] + system_above_bools[k] == 0: #both planets are below the boundary
                            D_ratio_below_cand.append(system_D_ratio[i_ratio])

    Rstar_sys_cand = np.array(Rstar_sys_cand)
    Mstar_sys_cand = np.array(Mstar_sys_cand)
    teff_sys_cand = np.array(teff_sys_cand)
    bp_rp_sys_cand = np.array(bp_rp_sys_cand)
    e_bp_rp_sys_cand = np.array(e_bp_rp_sys_cand)
    cdpp4p5_sys_cand = np.array(cdpp4p5_sys_cand)

    P_sys_cand = np.array(P_sys_cand)
    D_sys_cand = np.array(D_sys_cand)
    tdur_sys_cand = np.array(tdur_sys_cand)
    tdur_tcirc_sys_cand = np.array(tdur_tcirc_sys_cand)
    radii_sys_cand = np.array(radii_sys_cand)
    Rm_sys_cand = np.array(Rm_sys_cand)
    D_ratio_sys_cand = np.array(D_ratio_sys_cand)
    xi_sys_cand = np.array(xi_sys_cand)

    P_cand = planets_cleaned['P']
    M_cand = np.array(M_cand)
    Nmult_cand = np.array([np.sum(M_cand == x) for x in range(1,np.max(M_cand)+1)])
    Rm_cand = np.array(Rm_cand)
    D_ratio_cand = np.array(D_ratio_cand)
    xi_cand = np.array(xi_cand)
    xi_res_cand = np.array(xi_res_cand)
    xi_res32_cand = np.array(xi_res32_cand)
    xi_res21_cand = np.array(xi_res21_cand)
    xi_nonres_cand = np.array(xi_nonres_cand)

    D_above_cand = np.array(D_above_cand)
    D_below_cand = np.array(D_below_cand)
    D_ratio_above_cand = np.array(D_ratio_above_cand)
    D_ratio_below_cand = np.array(D_ratio_below_cand)
    D_ratio_across_cand = np.array(D_ratio_across_cand)

    tdur_tcirc_1_cand = tdur_tcirc_sys_cand[M_cand == 1, 0] # observed singles, 1d
    tdur_tcirc_2p_cand = tdur_tcirc_sys_cand[M_cand > 1] # observed multis, but 2d
    tdur_tcirc_2p_cand = tdur_tcirc_2p_cand[~np.isnan(tdur_tcirc_2p_cand)] # observed multis, 1d

    # Create dictionaries to hold summary stats ('ssk' stands for 'summary stats Kepler'):

    ssk_per_sys = {}
    # Stellar properties:
    ssk_per_sys['Rstar_obs'] = Rstar_sys_cand
    ssk_per_sys['Mstar_obs'] = Mstar_sys_cand
    ssk_per_sys['teff_obs'] = teff_sys_cand
    ssk_per_sys['bp_rp_obs'] = bp_rp_sys_cand
    ssk_per_sys['e_bp_rp_obs'] = e_bp_rp_sys_cand
    ssk_per_sys['cdpp4p5_obs'] = cdpp4p5_sys_cand
    # Planet properties:
    ssk_per_sys['Mtot_obs'] = M_cand
    ssk_per_sys['P_obs'] = P_sys_cand
    ssk_per_sys['D_obs'] = D_sys_cand
    ssk_per_sys['tdur_obs'] = tdur_sys_cand
    ssk_per_sys['tdur_tcirc_obs'] = tdur_tcirc_sys_cand
    ssk_per_sys['radii_obs'] = radii_sys_cand
    # Planet property ratios:
    ssk_per_sys['Rm_obs'] = Rm_sys_cand
    ssk_per_sys['D_ratio_obs'] = D_ratio_sys_cand
    ssk_per_sys['xi_obs'] = xi_sys_cand

    ssk = {}
    # Stellar properties (repeated to match number of planets):
    ssk['Rstar_obs'] = Rstar_cand
    ssk['Mstar_obs'] = Mstar_cand
    ssk['teff_obs'] = teff_cand
    ssk['bp_rp_obs'] = bp_rp_cand
    ssk['e_bp_rp_obs'] = e_bp_rp_cand
    ssk['cdpp4p5_obs'] = cdpp4p5_cand
    # Planet multiplicities:
    ssk['Nmult_obs'] = Nmult_cand
    # Planet properties:
    ssk['P_obs'] = P_cand
    ssk['D_obs'] = D_cand
    ssk['tdur_obs'] = tdur_cand
    ssk['tdur_tcirc_obs'] = tdur_tcirc_sys_cand[~np.isnan(tdur_tcirc_sys_cand)]
    ssk['tdur_tcirc_1_obs'] = tdur_tcirc_1_cand
    ssk['tdur_tcirc_2p_obs'] = tdur_tcirc_2p_cand
    ssk['radii_obs'] = radii_cand
    ssk['D_above_obs'] = D_above_cand
    ssk['D_below_obs'] = D_below_cand
    # Planet property ratios:
    ssk['Rm_obs'] = Rm_cand
    ssk['D_ratio_obs'] = D_ratio_cand
    ssk['xi_obs'] = xi_cand
    ssk['xi_res_obs'] = xi_res_cand
    ssk['xi_res32_obs'] = xi_res32_cand
    ssk['xi_res21_obs'] = xi_res21_cand
    ssk['xi_nonres_obs'] = xi_nonres_cand
    ssk['D_ratio_above_obs'] = D_ratio_above_cand
    ssk['D_ratio_below_obs'] = D_ratio_below_cand
    ssk['D_ratio_across_obs'] = D_ratio_across_cand

    # To compute some summary stats (system-level metrics) from GF2020:
    Nsys_obs = len(M_cand)
    assert Nsys_obs == len(radii_sys_cand) == len(P_sys_cand)

    radii_star_ratio = []
    radii_partitioning = []
    radii_monotonicity = []
    gap_complexity = []
    for i in range(Nsys_obs):
        P_sys = P_sys_cand[i][P_sys_cand[i] > 0]
        radii_sys = radii_sys_cand[i][P_sys_cand[i] > 0]

        radii_star_ratio.append(gen.radii_star_ratio(radii_sys, Rstar_sys_cand[i]))

        if len(radii_sys) >= 2:
            radii_partitioning.append(gen.partitioning(radii_sys))
            radii_monotonicity.append(gen.monotonicity_GF2020(radii_sys))
        if len(P_sys) >= 3:
            gap_complexity.append(gen.gap_complexity_GF2020(P_sys))
    ssk_per_sys['radii_star_ratio'] = np.array(radii_star_ratio)
    ssk_per_sys['radii_partitioning'] = np.array(radii_partitioning)
    ssk_per_sys['radii_monotonicity'] = np.array(radii_monotonicity)
    ssk_per_sys['gap_complexity'] = np.array(gap_complexity)

    return [ssk_per_sys, ssk]





# Functions for computing distances between the simulated and Kepler observed catalogs:

def CRPD_dist(En, On):
    """
    Compute the Cressie-Read Power Divergence (CRPD) statistic for observed planet multiplicity distributions.

    Warning
    -------
    Can potentially return negative values for extreme/edge cases!

    Parameters
    ----------
    En : array[int]
        The 'expected' (i.e. simulated) numbers of total systems with 1,2,3,... observed planets.
    On : array[int]
        The 'observed' (i.e. actual Kepler) numbers of total systems with 1,2,3,... observed planets.

    Returns
    -------
    rho : float
        The CRPD statistic.
    """
    n_max = max(len(En), len(On))
    En = np.array(list(En) + [0]*(n_max - len(En)))
    On = np.array(list(On) + [0]*(n_max - len(On)))

    E_array = En/float(np.sum(En)) # normalized numbers (fractions) of simulated systems with 1,2,3,... observed planets
    O_array = On/float(np.sum(On)) # normalized numbers (fractions) of actual Kepler systems with 1,2,3,... observed planets
    rho = 0.
    for i,E_i in enumerate(E_array):
        if En[i] != 0:
            rho += O_array[i]*((O_array[i]/E_array[i])**(2./3.) - 1.)
    rho = (9./5.)*rho

    return rho

def KS_dist_mult(x1, x2):
    """
    Compute the two-sample Kolmogorov-Smirnov (KS) distance between two discrete distributions taking on integer values.

    Parameters
    ----------
    x1, x2 : array[int]
        A sample of integers.

    Returns
    -------
    KS_dist : float
        The KS distance between the two distributions (i.e. the greatest distance between the cumulative distributions).
    KS_x : float
        The x-value that corresponds to the greatest distance between the two cumulative distributions.
    """
    x12_max = np.max((np.max(x1), np.max(x2))) # maximum value of x1 and x2
    x1_counts, x1_bins = np.histogram(x1, bins=x12_max, range=(0.5, x12_max+0.5))
    x2_counts, x2_bins = np.histogram(x2, bins=x12_max, range=(0.5, x12_max+0.5))
    pdf_diffs = x1_counts/np.float(len(x1)) - x2_counts/np.float(len(x2))
    cdf_diffs = np.cumsum(pdf_diffs)
    KS_dist = np.max(np.abs(cdf_diffs)) # K-S distance
    KS_x = np.arange(1, x12_max+1)[np.where(np.abs(cdf_diffs) == KS_dist)[0][0]] # x value where the CDF difference is the largest

    return KS_dist, KS_x

def KS_dist(x1, x2):
    """
    Compute the two-sample Kolmogorov-Smirnov (KS) distance between two continuous distributions (i.e. no repeat values).

    Parameters
    ----------
    x1, x2 : array[float]
        A sample of real values.

    Returns
    -------
    KS_dist : float
        The KS distance between the two distributions (i.e. the greatest distance between the cumulative distributions).
    KS_x : float
        The x-value that corresponds to the greatest distance between the two cumulative distributions.
    """
    x_all = np.concatenate((x1, x2)) # combined array
    i_all_sorted = np.argsort(x_all) # array of indices that would sort the combined array
    pdf_diffs = np.concatenate((np.ones(len(x1))/np.float(len(x1)), -np.ones(len(x2))/np.float(len(x2))))[i_all_sorted]
    cdf_diffs = np.cumsum(pdf_diffs)
    KS_dist = np.max(np.abs(cdf_diffs)) # K-S distance
    KS_x = x_all[i_all_sorted][np.where(np.abs(cdf_diffs) == KS_dist)[0][0]] # x value (a value in either x1 or x2) where the CDF difference is the largest

    return KS_dist, KS_x

def AD_dist(x1, x2):
    """
    Compute the two-sample Anderson-Darling (AD) distance between two continuous distributions.

    Implements Equation 1.2 of A. N. Pettitt (1976).

    Note
    ----
    Returns ``np.inf`` if there are not enough points (less than two) in either ``x1`` or ``x2`` for computing the AD distance.

    Parameters
    ----------
    x1, x2 : array[float]
        A sample of real values.

    Returns
    -------
    AD_dist : float
        The AD distance between the two distributions.
    """
    n, m = len(x1), len(x2)
    if n > 1 and m > 1:
        N = n + m
        x_all = np.concatenate((x1, x2)) # combined array
        i_all_sorted = np.argsort(x_all) # array of indices that would sort the combined array
        M_i_diffs = np.concatenate((np.ones(n), np.zeros(m)))[i_all_sorted]
        M_i_array = np.cumsum(M_i_diffs)[:-1] # array of M_i except for last element, i.e. from i=1 to i=N-1
        i_array = 1. + np.arange(N-1) # array of i from i=1 to i=N-1
        AD_dist = (1./(n*m))*np.sum(((M_i_array*N - n*i_array)**2.)/(i_array*(N - i_array))) # AD distance
    else:
        print('Not enough points to compute AD distance; returning inf.')
        AD_dist = np.inf

    return AD_dist

def AD_dist2(x1, x2):
    """
    Compute the two-sample Anderson-Darling (AD) distance between two continuous distributions.

    Implements Equation 3 of Scholz \& Stephens (1987). Tested to be equivalent to :py:func:`syssimpyplots.compare_kepler.AD_dist`.

    Note
    ----
    Returns ``np.inf`` if there are not enough points (less than two) in either ``x1`` or ``x2`` for computing the AD distance.

    Parameters
    ----------
    x1, x2 : array[float]
        A sample of real values.

    Returns
    -------
    AD_dist : float
        The AD distance between the two distributions.
    """
    n1, n2 = len(x1), len(x2)
    if n1 > 1 and n2 > 1:
        N = n1 + n2
        x_all = np.concatenate((x1, x2)) # combined array
        i_all_sorted = np.argsort(x_all) # array of indices that would sort the combined array

        M_1j_diffs = np.concatenate((np.ones(n1), np.zeros(n2)))[i_all_sorted]
        M_1j_array = np.cumsum(M_1j_diffs)[:-1] # array of M_1j except for last element, i.e. from j=1 to j=N-1
        M_2j_diffs = np.concatenate((np.zeros(n1), np.ones(n2)))[i_all_sorted]
        M_2j_array = np.cumsum(M_2j_diffs)[:-1] # array of M_2j except for last element, i.e. from j=1 to j=N-1
        j_array = 1. + np.arange(N-1) # array of j from j=1 to j=N-1

        AD_dist = (1./N)*((1./n1)*np.sum(((N*M_1j_array - n1*j_array)**2.)/(j_array*(N - j_array))) + (1./n2)*np.sum(((N*M_2j_array - n2*j_array)**2.)/(j_array*(N - j_array)))) # AD distance
    else:
        print('Not enough points to compute AD distance; returning inf.')
        AD_dist = np.inf

    return AD_dist

def AD_mod_dist(x1, x2):
    """
    Compute a modified version of the two-sample Anderson-Darling (AD) distance between two continuous distributions.

    Equivalent to the AD distance (implemented by :py:func:`syssimpyplots.compare_kepler.AD_dist` and :py:func:`syssimpyplots.compare_kepler.AD_dist2`) without the factor of 'n*m/N' in front of the integral, where 'n' and 'm' are the sample sizes (and 'N=n+m' is the combined sample size).

    Note
    ----
    Returns ``np.inf`` if there are not enough points (less than two) in either ``x1`` or ``x2`` for computing the AD distance.

    Parameters
    ----------
    x1, x2 : array[float]
        A sample of real values.

    Returns
    -------
    AD_dist : float
        The (modified) AD distance between the two distributions.
    """
    n, m = len(x1), len(x2)
    if n > 1 and m > 1:
        N = n + m
        x_all = np.concatenate((x1, x2)) # combined array
        i_all_sorted = np.argsort(x_all) # array of indices that would sort the combined array
        M_i_diffs = np.concatenate((np.ones(n), np.zeros(m)))[i_all_sorted]
        M_i_array = np.cumsum(M_i_diffs)[:-1] # array of M_i except for last element, i.e. from i=1 to i=N-1
        i_array = 1. + np.arange(N-1) # array of i from i=1 to i=N-1
        AD_dist = (N/((n*m)**2.))*np.sum(((M_i_array*N - n*i_array)**2.)/(i_array*(N - i_array))) # AD distance
    else:
        print('Not enough points to compute AD distance; returning inf.')
        AD_dist = np.inf

    return AD_dist

def load_split_stars_model_evaluations_and_weights(file_name):
    """
    Load a file containing the distances from many evaluations of the same model, and compute the weights for each distance term.

    Parameters
    ----------
    file_name : str
        The path/name of the file containing the distances of many model evaluations.

    Returns
    -------
    Nmult_evals : dict
        A dictionary containing an array of observed planet multiplicity distributions for each model evaluation, for each stellar sample (`all`, `bluer`, and `redder` fields).
    d_all_keys_evals : dict
        A dictionary containing an array of distance term names (strings) for each model evaluation, for each stellar sample.
    d_all_vals_evals : dict
        A dictionary containing an array of distances (corresponding to the distance term names) for each model evaluation, for each stellar sample.
    weights_all : dict
        A dictionary containing a dictionary for the weights corresponding to each distance term, for each stellar sample.


    Note
    ----
    The `bluer` and `redder` samples split the stellar sample into two equal sized samples of stars below and above the median Gaia DR2 bp-rp color, respectively.

    Warning
    -------
    Currently returns empty arrays in the `Nmult_evals` dictionary.
    """
    sample_names = ['all', 'bluer', 'redder']

    Nmult_max = 8
    Nmult_evals = {key: [] for key in sample_names}
    d_all_keys_evals = {key: [] for key in sample_names}
    d_all_vals_evals = {key: [] for key in sample_names}

    with open(file_name, 'r') as file:
        for line in file:
            for key in sample_names:
                n = len(key)
                if line[0:n+2] == '[%s]' % key:
                    #if line[n+3:n+3+6] == 'Counts':
                    #Nmult_str, counts_str = line[n+3+9:-2].split('][')
                    #Nmult = tuple([int(x) for x in Nmult_str.split(', ')])
                    #Nmult_evals[key].append(Nmult)

                    if line[n+3:n+3+11] == 'd_all_keys:':
                        d_all_keys = line[n+3+14:-3].split('", "')
                        d_all_keys_evals[key].append(d_all_keys)

                    elif line[n+3:n+3+11] == 'd_all_vals:':
                        d_all_vals_str = line[n+3+13:-2].split(', ')
                        d_all_vals = tuple([float(x) for x in d_all_vals_str])
                        d_all_vals_evals[key].append(d_all_vals)

    for key in sample_names:
        Nmult_evals[key] = np.array(Nmult_evals[key], dtype=[(str(n), 'i8') for n in range(1,Nmult_max+1)])
        d_all_keys_evals[key] = np.array(d_all_keys_evals[key])
        d_all_vals_evals[key] = np.array(d_all_vals_evals[key], dtype=[(dist_key, 'f8') for dist_key in d_all_keys_evals[key][0]])

    weights_all = {}
    for key in sample_names:
        dict_weights = {}
        for dist_key in d_all_vals_evals[key].dtype.names:
            dict_weights[dist_key] = 1./np.sqrt(np.mean(d_all_vals_evals[key][dist_key]**2.))
        weights_all[key] = dict_weights

    return Nmult_evals, d_all_keys_evals, d_all_vals_evals, weights_all

def load_split_stars_weights_only():
    """
    Compute the weights for each distance term.

    Wrapper to return just the weights from the function :py:func:`syssimpyplots.compare_kepler.load_split_stars_model_evaluations_and_weights`.
    """
    Nmult_evals, d_all_keys_evals, d_all_vals_evals, weights_all = load_split_stars_model_evaluations_and_weights(os.path.join(path_data, 'Clustered_P_R_split_stars_weights_ADmod_true_targs88912_evals100_all_pairs.txt'))
    return weights_all

def compute_total_weighted_dist(weights, dists, dists_w, dists_include=[]):
    """
    Compute the total weighted distance including a number of distance terms.

    Also prints out the individual distance terms, their weights, and their unweighted and weighted distances.

    Parameters
    ----------
    weights : dict
        The dictionary containing the weights for to each distance term.
    dists : dict
        The dictionary containing the individual distance terms.
    dists_w : dict
        The dictionary containing the individual weighted distance terms.
    dists_include : list[str], default=[]
        The list of distance terms (strings) to include in the sum.

    Returns
    -------
    tot_dist_w : float
        The total weighted distance of the included distance terms.
    """
    if len(dists_include) == 0:
        print('No distance terms to include.')

    print('#####')
    tot_dist_w = 0.
    for key in dists_include:
        tot_dist_w += dists_w[key]
        print('{:<30}: weight = {:<8}, dist = {:<8}, dist_w = {:<8}'.format(key, np.round(weights[key],2), np.round(dists[key],4), np.round(dists_w[key],2)))
    print('Total weighted distance = %s' % tot_dist_w)
    print('#####')

    return tot_dist_w

def compute_distances_sim_Kepler(sss_per_sys, sss, ssk_per_sys, ssk, weights, dists_include, N_sim, cos_factor=1., AD_mod=True, print_dists=True):
    """
    Compute weighted and unweighted distances for a large collection of distance terms.

    Parameters
    ----------
    sss_per_sys : dict
        A dictionary of summary statistics per observed system in a simulated observed catalog.
    sss : dict
        A dictionary of summary statistics for all observed planets in a simulated observed catalog.
    ssk_per_sys : dict
        A dictionary of summary statistics per observed system in the Kepler catalog.
    ssk : dict
        A dictionary of summary statistics for observed all planets in the Kepler catalog.
    weights : dict
        A dictionary of the weights corresponding to each distance term.
    dists_include : list[str]
        A list of distance terms (strings) to be printed.
    N_sim : int
        The number of target stars (i.e. simulated systems) in the simulated catalog.
    cos_factor : float, default=1.
        The cosine of the maximum inclination angle (relative to the sky plane) drawn for the reference planes of the simulated systems (between 0 and 1).
    AD_mod : bool, default=True
        Whether to compute the modified AD distance (:py:func:`syssimpyplots.compare_kepler.AD_mod_dist`, if True) or the standard AD distance (:py:func:`syssimpyplots.compare_kepler.AD_dist`, if False).
    print_dists : bool, default=True
        Whether to print the distances corresponding to the terms in `dists_include`. If True, also prints the total numbers of observed planets and planet pairs in the simulated and Kepler catalogs.

    Returns
    -------
    dists : dict
        A dictionary containing all the various (unweighted) distance terms.
    dists_w : dict
        A dictionary containing all the various (weighted) distance terms.


    Note
    ----
    The distance terms computed and included in ``dists`` and ``dists_w`` are not limited to the terms in ``dists_include``.
    """
    # Create a dict to hold all distance terms:
    dists = {}

    dists['delta_f'] = np.abs(len(sss['P_obs'])/(float(N_sim)/cos_factor) - len(ssk['P_obs'])/float(N_Kep)) # absolute difference in the rates of observed planets per star

    Nmult_obs_sim_5plus = np.array(list(sss['Nmult_obs'][:4]) + [sum(sss['Nmult_obs'][4:])])
    Nmult_obs_Kep_5plus = np.array(list(ssk['Nmult_obs'][:4]) + [sum(ssk['Nmult_obs'][4:])])
    dists['mult_CRPD'] = CRPD_dist(Nmult_obs_sim_5plus, Nmult_obs_Kep_5plus)
    dists['mult_CRPD_r'] = CRPD_dist(Nmult_obs_Kep_5plus, Nmult_obs_sim_5plus)

    R_res32_sim, R_res32_Kep = np.float(sum((sss['Rm_obs'] >= 1.5) & (sss['Rm_obs'] <= 1.5*(1.+gen.res_width))))/np.float(len(sss['Rm_obs'])), np.float(sum((ssk['Rm_obs'] >= 1.5) & (ssk['Rm_obs'] <= 1.5*(1.+gen.res_width))))/np.float(len(ssk['Rm_obs'])) # fractions of planet pairs within 5% of 3:2 MMR, for simulated and Kepler data
    R_res21_sim, R_res21_Kep = np.float(sum((sss['Rm_obs'] >= 2.) & (sss['Rm_obs'] <= 2.*(1.+gen.res_width))))/np.float(len(sss['Rm_obs'])), np.float(sum((ssk['Rm_obs'] >= 2.) & (ssk['Rm_obs'] <= 2.*(1.+gen.res_width))))/np.float(len(ssk['Rm_obs'])) # fractions of planet pairs within 5% of 2:1 MMR, for simulated and Kepler data
    R_res32_diff = np.abs(R_res32_sim - R_res32_Kep) # difference in fractions of planet pairs close to 3:2 MMR between simulated and Kepler data
    R_res21_diff = np.abs(R_res21_sim - R_res21_Kep) # difference in fractions of planet pairs close to 2:1 MMR between simulated and Kepler data

    tdur_tcirc_1 = sss_per_sys['tdur_tcirc_obs'][sss_per_sys['Mtot_obs'] == 1, 0]
    tdur_tcirc_Kep_1 = ssk_per_sys['tdur_tcirc_obs'][ssk_per_sys['Mtot_obs'] == 1, 0]
    tdur_tcirc_2p = sss_per_sys['tdur_tcirc_obs'][sss_per_sys['Mtot_obs'] > 1]
    tdur_tcirc_2p = tdur_tcirc_2p[tdur_tcirc_2p > 0]
    tdur_tcirc_Kep_2p = ssk_per_sys['tdur_tcirc_obs'][ssk_per_sys['Mtot_obs'] > 1]
    tdur_tcirc_Kep_2p = tdur_tcirc_Kep_2p[tdur_tcirc_Kep_2p > 0]

    # KS distances:
    dists['periods_KS'] = KS_dist(sss['P_obs'], ssk['P_obs'])[0]
    dists['period_ratios_KS'] = KS_dist(sss['Rm_obs'], ssk['Rm_obs'])[0]
    dists['durations_KS'] = KS_dist(sss['tdur_obs'], ssk['tdur_obs'])[0]
    dists['durations_norm_circ_KS'] = KS_dist(sss['tdur_tcirc_obs'], ssk['tdur_tcirc_obs'])[0]
    dists['durations_norm_circ_singles_KS'] = KS_dist(tdur_tcirc_1, tdur_tcirc_Kep_1)[0]
    dists['durations_norm_circ_multis_KS'] = KS_dist(tdur_tcirc_2p, tdur_tcirc_Kep_2p)[0]
    dists['duration_ratios_KS'] = KS_dist(sss['xi_obs'], ssk['xi_obs'])[0]
    dists['duration_ratios_nonmmr_KS'] = KS_dist(sss['xi_nonres_obs'], ssk['xi_nonres_obs'])[0]
    dists['duration_ratios_mmr_KS'] = KS_dist(sss['xi_res_obs'], ssk['xi_res_obs'])[0]
    dists['depths_KS'] = KS_dist(sss['D_obs'], ssk['D_obs'])[0]
    dists['depths_above_KS'] = KS_dist(sss['D_above_obs'], ssk['D_above_obs'])[0]
    dists['depths_below_KS'] = KS_dist(sss['D_below_obs'], ssk['D_below_obs'])[0]
    dists['radius_ratios_KS'] = KS_dist(sss['D_ratio_obs'], ssk['D_ratio_obs'])[0]
    dists['radius_ratios_above_KS'] = KS_dist(sss['D_ratio_above_obs'], ssk['D_ratio_above_obs'])[0]
    dists['radius_ratios_below_KS'] = KS_dist(sss['D_ratio_below_obs'], ssk['D_ratio_below_obs'])[0]
    dists['radius_ratios_across_KS'] = KS_dist(sss['D_ratio_across_obs'], ssk['D_ratio_across_obs'])[0]
    dists['radii_partitioning_KS'] = KS_dist(sss_per_sys['radii_partitioning'], ssk_per_sys['radii_partitioning'])[0]
    dists['radii_monotonicity_KS'] = KS_dist(sss_per_sys['radii_monotonicity'], ssk_per_sys['radii_monotonicity'])[0]
    dists['gap_complexity_KS'] = KS_dist(sss_per_sys['gap_complexity'], ssk_per_sys['gap_complexity'])[0]

    # AD distances:
    AD_stat = AD_mod_dist if AD_mod else AD_dist

    dists['periods_AD'] = AD_stat(sss['P_obs'], ssk['P_obs'])
    dists['period_ratios_AD'] = AD_stat(sss['Rm_obs'], ssk['Rm_obs'])
    dists['durations_AD'] = AD_stat(sss['tdur_obs'], ssk['tdur_obs'])
    dists['durations_norm_circ_AD'] = AD_stat(sss['tdur_tcirc_obs'], ssk['tdur_tcirc_obs'])
    dists['durations_norm_circ_singles_AD'] = AD_stat(tdur_tcirc_1, tdur_tcirc_Kep_1)
    dists['durations_norm_circ_multis_AD'] = AD_stat(tdur_tcirc_2p, tdur_tcirc_Kep_2p)
    dists['duration_ratios_AD'] = AD_stat(sss['xi_obs'], ssk['xi_obs'])
    dists['duration_ratios_nonmmr_AD'] = AD_stat(sss['xi_nonres_obs'], ssk['xi_nonres_obs'])
    dists['duration_ratios_mmr_AD'] = AD_stat(sss['xi_res_obs'], ssk['xi_res_obs'])
    dists['depths_AD'] = AD_stat(sss['D_obs'], ssk['D_obs'])
    dists['depths_above_AD'] = AD_stat(sss['D_above_obs'], ssk['D_above_obs'])
    dists['depths_below_AD'] = AD_stat(sss['D_below_obs'], ssk['D_below_obs'])
    dists['radius_ratios_AD'] = AD_stat(sss['D_ratio_obs'], ssk['D_ratio_obs'])
    dists['radius_ratios_above_AD'] = AD_stat(sss['D_ratio_above_obs'], ssk['D_ratio_above_obs'])
    dists['radius_ratios_below_AD'] = AD_stat(sss['D_ratio_below_obs'], ssk['D_ratio_below_obs'])
    dists['radius_ratios_across_AD'] = AD_stat(sss['D_ratio_across_obs'], ssk['D_ratio_across_obs'])
    dists['radii_partitioning_AD'] = AD_stat(sss_per_sys['radii_partitioning'], ssk_per_sys['radii_partitioning'])
    dists['radii_monotonicity_AD'] = AD_stat(sss_per_sys['radii_monotonicity'], ssk_per_sys['radii_monotonicity'])
    dists['gap_complexity_AD'] = AD_stat(sss_per_sys['gap_complexity'], ssk_per_sys['gap_complexity'])

    # To create a dict for all the weighted distances:
    dists_w = {}
    for key in dists.keys():
        dists_w[key] = dists[key]*weights[key]

    # To compute the total weighted KS and AD distances assuming default terms:
    dists_default = ['periods', 'period_ratios', 'durations', 'duration_ratios_nonmmr', 'duration_ratios_mmr', 'depths', 'radius_ratios']
    tot_dist_w_KS = dists_w['delta_f'] + dists_w['mult_CRPD_r']
    tot_dist_w_AD = dists_w['delta_f'] + dists_w['mult_CRPD_r']
    for key in dists_default:
        tot_dist_w_KS += dists_w[key+'_KS']
        tot_dist_w_AD += dists_w[key+'_AD']
    dists_w['tot_dist_KS_default'] = tot_dist_w_KS
    dists_w['tot_dist_AD_default'] = tot_dist_w_AD



    if print_dists:
        print('(Planets Kepler obs, Planet pairs Kepler obs) = (%s, %s)' % (len(ssk['P_obs']), len(ssk['Rm_obs'])))
        print('(Planets obs, Planet pairs obs) = (%s, %s)' % (len(sss['P_obs']), len(sss['Rm_obs'])))
        tot_dist_w = compute_total_weighted_dist(weights, dists, dists_w, dists_include=dists_include)
        dists_w['tot_dist_w_include'] = tot_dist_w

    return dists, dists_w



# To import required modules:
import numpy as np
import os
import sys
from itertools import chain # for rapid flattening of lists of lists

from syssimpyplots.general import *
from syssimpyplots.compare_kepler import *
from syssimpyplots.load_sims import *





# Test functions in "src/functions_load_sims.py":

loadfiles_directory = '/Users/hematthi/Documents/GradSchool/Research/ACI/Simulated_Data/AMD_system/Split_stars/Singles_ecc/Params11_KS/Distribute_AMD_per_mass/durations_norm_circ_singles_multis_GF2020_KS/GP_med/'

loadfiles_directory_multiple = '/Users/hematthi/Documents/GradSchool/Research/ACI/Simulated_Data/AMD_system/Split_stars/Singles_ecc/Params11_KS/Distribute_AMD_per_mass/durations_norm_circ_singles_multis_GF2020_KS/GP_best_models/'

def test_read_targets_period_radius_bounds(load_dir=loadfiles_directory, run_number=''):
    N_sim, cos_factor, P_min, P_max, radii_min, radii_max = read_targets_period_radius_bounds(load_dir + 'periods%s.out' % run_number)
    assert type(N_sim) is int
    assert 0 <= cos_factor <= 1
    assert 3. <= P_min < P_max <= 300.
    assert 0.5 <= radii_min < radii_max <= 10.

def test_load_cat_phys(load_dir=loadfiles_directory, run_number=''):
    cat_phys = load_cat_phys(load_dir + 'physical_catalog.csv')
    N_sim, cos_factor, P_min, P_max, radii_min, radii_max = read_targets_period_radius_bounds(load_dir + 'physical_catalog%s.csv' % run_number)
    assert 0 <= np.min(cat_phys['planet_mass'])
    assert radii_min <= np.min(cat_phys['planet_radius']) * Rsun/Rearth
    #assert radii_max >= np.max(cat_phys['planet_radius']) * Rsun/Rearth # NOTE: fails due to differences in precision of Rsun/Rearth in SysSim vs. in 'functions_general.py'
    assert P_min <= np.min(cat_phys['period']) <= np.max(cat_phys['period']) <= P_max
    assert 0 <= np.min(cat_phys['ecc']) <= np.max(cat_phys['ecc']) <= 1
    assert 0 <= np.min(cat_phys['incl']) <= np.max(cat_phys['incl']) <= np.pi
    assert -np.pi <= np.min(cat_phys['omega']) <= np.max(cat_phys['omega']) <= np.pi
    assert 0 <= np.min(cat_phys['asc_node']) <= np.max(cat_phys['asc_node']) <= 2.*np.pi
    assert 0 <= np.min(cat_phys['mean_anom']) <= np.max(cat_phys['mean_anom']) <= 2.*np.pi
    assert 0 <= np.min(cat_phys['incl_invariable']) <= np.max(cat_phys['incl_invariable']) <= np.pi
    assert 0 <= np.min(cat_phys['asc_node_invariable']) <= np.max(cat_phys['asc_node_invariable']) <= 2.*np.pi
    assert 0 < np.min(cat_phys['star_mass'])
    assert 0 < np.min(cat_phys['star_radius'])

def test_load_star_phys(load_dir=loadfiles_directory, run_number=''):
    star_phys = load_star_phys(load_dir + 'physical_catalog_stars%s.csv' % run_number)
    assert 0 < np.min(star_phys['star_mass']) <= np.max(star_phys['star_mass']) < 2.
    assert 0 < np.min(star_phys['star_radius']) <= np.max(star_phys['star_radius']) < 4.
    assert 0 <= np.min(star_phys['num_planets'])

def test_load_planets_stars_phys_separate(load_dir=loadfiles_directory, run_number=''):
    clusterids_per_sys, P_per_sys, radii_per_sys, mass_per_sys, e_per_sys, inclmut_per_sys, incl_per_sys, Mstar_all, Rstar_all = load_planets_stars_phys_separate(load_dir, run_number=run_number)
    N_sim, cos_factor, P_min, P_max, radii_min, radii_max = read_targets_period_radius_bounds(load_dir + 'periods_all%s.out' % run_number)

    clusterids = list(chain(*clusterids_per_sys))
    periods = list(chain(*P_per_sys))
    radii = list(chain(*radii_per_sys))
    masses = list(chain(*mass_per_sys))
    eccs = list(chain(*e_per_sys))
    inclmuts = list(chain(*inclmut_per_sys))
    incls = list(chain(*incl_per_sys))
    N_sys_pl, N_pl = len(clusterids_per_sys), len(clusterids)

    assert N_pl == len(periods) == len(radii) == len(masses) == len(eccs) == len(inclmuts) == len(incls)
    assert N_sys_pl == len(Mstar_all) == len(Rstar_all) <= N_sim
    assert P_min <= np.min(periods) <= np.max(periods) <= P_max
    assert radii_min <= np.min(radii) * Rsun/Rearth
    #assert radii_max >= np.max(radii) * Rsun/Rearth # NOTE: fails due to differences in precision of Rsun/Rearth in SysSim vs. in 'functions_general.py'
    assert 0 <= np.min(masses)
    assert 0 <= np.min(eccs) <= np.max(eccs) <= 1
    assert 0 <= np.min(inclmuts) <= np.max(inclmuts) <= np.pi
    assert 0 <= np.min(incls) <= np.max(incls) <= np.pi
    assert 0 < np.min(Mstar_all)
    assert 0 < np.min(Rstar_all)

def test_compute_basic_summary_stats_per_sys_cat_phys(load_dir=loadfiles_directory, run_number=''):
    clusterids_per_sys, P_per_sys, radii_per_sys, mass_per_sys, e_per_sys, inclmut_per_sys, incl_per_sys, Mstar_all, Rstar_all = load_planets_stars_phys_separate(load_dir, run_number=run_number)
    N_sim, cos_factor, P_min, P_max, radii_min, radii_max = read_targets_period_radius_bounds(load_dir + 'periods_all%s.out' % run_number)

    N_sys_pl, N_pl = len(clusterids_per_sys), len(list(chain(*clusterids_per_sys)))

    sssp_per_sys_basic = compute_basic_summary_stats_per_sys_cat_phys(clusterids_per_sys, P_per_sys, radii_per_sys, mass_per_sys, e_per_sys, inclmut_per_sys, incl_per_sys, Mstar_all, Rstar_all)
    assert type(sssp_per_sys_basic['Mmax']) is np.int64
    assert sssp_per_sys_basic['Mmax'] == np.max(sssp_per_sys_basic['Mtot_all'])
    assert N_sys_pl == len(sssp_per_sys_basic['Mtot_all'])
    assert N_pl == np.sum(sssp_per_sys_basic['Mtot_all']) == np.sum(sssp_per_sys_basic['pl_per_cluster_all'])

    # Check that all fields have the right number of systems:
    keys = ['P_all', 'clusterids_all', 'e_all', 'inclmut_all', 'incl_all', 'radii_all', 'mass_all', 'Mstar_all', 'Rstar_all', 'mu_all', 'a_all', 'AMD_all', 'AMD_tot_all']
    for key in keys:
        assert N_sys_pl == len(sssp_per_sys_basic[key])

N_sim, cos_factor, P_min, P_max, radii_min, radii_max = read_targets_period_radius_bounds(loadfiles_directory + 'periods_all.out')

def run_test_summary_stats_cat_phys(sssp_per_sys, sssp, N_sim=N_sim, P_min=P_min, P_max=P_max, radii_min=radii_min, radii_max=radii_max):
    N_sys_pl = len(sssp_per_sys['Mtot_all']) # total number of systems
    N_pl = np.sum(sssp_per_sys['Mtot_all']) # total number of planets
    N_pl_pairs = np.sum(sssp_per_sys['Mtot_all'][sssp_per_sys['Mtot_all'] >= 2] - 1) # total number of adjacent pairs of planets

    # Check that all fields in 'sssp_per_sys' have the right number of systems:
    keys = ['clusterids_all', 'P_all', 'a_all', 'radii_all', 'mass_all', 'mu_all', 'e_all', 'inclmut_all', 'incl_all', 'AMD_all', 'Rm_all', 'radii_ratio_all', 'N_mH_all', 'dynamical_mass']
    for key in keys:
        assert N_sim == len(sssp_per_sys[key])
    assert len(sssp_per_sys['radii_partitioning']) == np.sum(sssp_per_sys['Mtot_all'] >= 2)
    assert len(sssp_per_sys['radii_monotonicity']) == np.sum(sssp_per_sys['Mtot_all'] >= 2)
    assert len(sssp_per_sys['gap_complexity']) == np.sum(sssp_per_sys['Mtot_all'] >= 3)

    # Check that all fields in 'sssp' have the right number of systems and planets:
    assert np.sum(sssp['clustertot_all']) == len(sssp['pl_per_cluster_all'])
    assert N_pl == np.sum(sssp['pl_per_cluster_all'])

    keys = ['Mstar_all', 'Rstar_all', 'clustertot_all', 'AMD_tot_all']
    for key in keys:
        assert N_sim == len(sssp[key])

    keys = ['P_all', 'radii_all', 'mass_all', 'e_all', 'incl_all']
    for key in keys:
        assert N_pl == len(sssp[key])
    assert np.sum(sssp_per_sys['Mtot_all'][sssp_per_sys['Mtot_all'] >= 2]) == len(sssp['inclmut_all'])
    assert N_pl == len(sssp['radii_above_all']) + len(sssp['radii_below_all'])
    assert N_pl_pairs == len(sssp['Rm_all'])
    assert N_pl_pairs == len(sssp['radii_ratio_all'])
    assert N_pl_pairs == len(sssp['N_mH_all'])
    assert N_pl_pairs == len(sssp['radii_ratio_above_all']) + len(sssp['radii_ratio_below_all']) + len(sssp['radii_ratio_across_all'])

    # Check that each field in 'sssp' has reasonable values:
    assert 0 < np.min(sssp['Mstar_all'])
    assert 0 < np.min(sssp['Rstar_all'])
    assert 0 <= np.min(sssp['clustertot_all'])
    assert 0 <= np.min(sssp['AMD_tot_all'])
    assert 1 <= np.min(sssp['pl_per_cluster_all'])
    assert P_min <= np.min(sssp['P_all']) <= np.max(sssp['P_all']) <= P_max
    assert radii_min <= np.min(sssp['radii_all']) <= np.max(sssp['radii_all']) #<= radii_max # NOTE: 'radii_max' check fails due to differences in precision of Rsun/Rearth in SysSim vs. in 'functions_general.py'
    assert 0 <= np.min(sssp['mass_all'])
    assert 0 <= np.min(sssp['e_all']) <= np.max(sssp['e_all']) <= 1
    assert 0 <= np.min(sssp['inclmut_all']) <= np.max(sssp['inclmut_all']) <= np.pi
    assert 0 <= np.min(sssp['incl_all']) <= np.max(sssp['incl_all']) <= np.pi
    assert radii_min <= np.min(sssp['radii_above_all']) <= np.max(sssp['radii_above_all']) #<= radii_max
    assert radii_min <= np.min(sssp['radii_below_all']) <= np.max(sssp['radii_below_all']) #<= radii_max
    assert 1 < np.min(sssp['Rm_all']) <= np.max(sssp['Rm_all']) <= P_max/P_min
    assert radii_min/radii_max <= np.min(sssp['radii_ratio_all']) <= np.max(sssp['radii_ratio_all']) <= radii_max/radii_min
    assert 0 < np.min(sssp['N_mH_all'])
    assert radii_min/radii_max <= np.min(sssp['radii_ratio_above_all']) <= np.max(sssp['radii_ratio_above_all']) <= radii_max/radii_min
    assert radii_min/radii_max <= np.min(sssp['radii_ratio_below_all']) <= np.max(sssp['radii_ratio_below_all']) <= radii_max/radii_min
    assert radii_min/radii_max <= np.min(sssp['radii_ratio_across_all']) <= np.max(sssp['radii_ratio_across_all']) <= radii_max/radii_min

def test_compute_summary_stats_from_cat_phys(load_dir=loadfiles_directory, run_number=''):
    N_sim, cos_factor, P_min, P_max, radii_min, radii_max = read_targets_period_radius_bounds(load_dir + 'periods_all%s.out' % run_number)
    sssp_per_sys, sssp = compute_summary_stats_from_cat_phys(file_name_path=load_dir, run_number=run_number, load_full_tables=True, match_observed=True)
    run_test_summary_stats_cat_phys(sssp_per_sys, sssp, N_sim=N_sim, P_min=P_min, P_max=P_max, radii_min=radii_min, radii_max=radii_max)

def test_load_cat_obs(load_dir=loadfiles_directory, run_number=''):
    cat_obs = load_cat_obs(load_dir + 'observed_catalog%s.csv' % run_number)
    assert 1 <= np.min(cat_obs['target_id']) <= np.max(cat_obs['target_id']) <= N_sim
    assert 1 <= np.min(cat_obs['star_id']) <= np.max(cat_obs['star_id']) <= N_Kep
    assert 0 < np.min(cat_obs['period']) # due to simulated observation uncertainty, can be slightly outside [P_min,P_max]
    assert 0 <= np.min(cat_obs['period_err']) <= np.max(cat_obs['period_err']) <= 0.1 # check range of period errors (days)
    assert 0 <= np.min(cat_obs['depth']) <= np.max(cat_obs['depth']) <= 1. # check range of transit depths
    assert 0 <= np.min(cat_obs['depth_err']) <= np.max(cat_obs['depth_err']) <= 1. # check range of transit depth errors
    assert 0 <= np.min(cat_obs['duration']) <= np.max(cat_obs['duration']) <= 1. # check range of transit durations (days)
    assert 0 <= np.min(cat_obs['duration_err']) <= np.max(cat_obs['duration_err']) <= 1. # check range of transit duration errors (days)
    assert 0 < np.min(cat_obs['star_mass'])
    assert 0 < np.min(cat_obs['star_radius'])

def test_load_star_obs(load_dir=loadfiles_directory, run_number=''):
    star_obs = load_star_obs(load_dir + 'observed_catalog_stars%s.csv' % run_number)
    assert 1 <= np.min(star_obs['target_id']) <= np.max(star_obs['target_id']) <= N_sim
    assert 1 <= np.min(star_obs['star_id']) <= np.max(star_obs['star_id']) <= N_Kep
    assert 0 < np.min(star_obs['star_mass']) <= np.max(star_obs['star_mass']) < 2. # check range of stellar masses (Solar masses)
    assert 0 < np.min(star_obs['star_radius']) <= np.max(star_obs['star_radius']) < 4. # check range of stellar radii (Solar radii)
    assert 1 <= np.min(star_obs['num_obs_planets']) <= np.max(star_obs['num_obs_planets']) <= 8 # NOTE: SysSim includes parameter 'max_tranets_in_sys' which is set to 8

def test_load_planets_stars_obs_separate(load_dir=loadfiles_directory, run_number=''):
    P_per_sys, D_per_sys, tdur_per_sys, Mstar_per_sys, Rstar_per_sys = load_planets_stars_obs_separate(load_dir,  run_number)

    periods = list(chain(*P_per_sys))
    depths = list(chain(*D_per_sys))
    durations = list(chain(*tdur_per_sys))
    N_sys, N_pl = len(P_per_sys), len(periods)

    assert N_sys == len(D_per_sys) == len(tdur_per_sys) == len(Mstar_per_sys) == len(Rstar_per_sys)
    assert N_pl == len(depths) == len(durations)
    assert 0 < np.min(periods) # due to simulated observation uncertainty, can be slightly outside [P_min,P_max]
    assert 0 <= np.min(depths) <= np.max(depths) <= 1. # check range of transit depths
    assert 0 <= np.min(durations) <= np.max(durations) <= 1. # check range of transit durations (days)
    assert 0 < np.min(Mstar_per_sys)
    assert 0 < np.min(Rstar_per_sys)

def run_test_summary_stats_cat_obs(sss_per_sys, sss):
    N_sys = len(sss_per_sys['Mstar_obs']) # total number of systems
    N_multis = np.sum(sss['Nmult_obs'][1:]) # total number of multi-planet systems
    assert N_sys == np.sum(sss['Nmult_obs'])
    N_pl = np.sum(sss_per_sys['Mtot_obs'])
    assert N_pl == np.sum(sss['Nmult_obs'] * np.arange(1,len(sss['Nmult_obs'])+1)) # total number of planets
    N_pl_pairs = np.sum(sss['Nmult_obs'] * np.arange(len(sss['Nmult_obs']))) # total number of adjacent planet pairs

    # Check that all fields in 'sss_per_sys' have the same number of systems:
    keys = ['Rstar_obs', 'Mstar_obs', 'teff_obs', 'bp_rp_obs', 'e_bp_rp_obs', 'cdpp4p5_obs', 'P_obs', 'D_obs', 'tdur_obs', 'tdur_tcirc_obs', 'radii_obs', 'Rm_obs', 'D_ratio_obs', 'xi_obs', 'xi_res_obs', 'xi_res32_obs', 'xi_res21_obs', 'xi_nonres_obs', 'radii_star_ratio']
    for key in keys:
        assert N_sys == len(sss_per_sys[key])

    assert N_multis == len(sss_per_sys['radii_partitioning'])
    assert N_multis == len(sss_per_sys['radii_monotonicity'])
    assert np.sum(sss['Nmult_obs'][2:]) == len(sss_per_sys['gap_complexity'])

    # Check that all fields in 'sss_per_sys' have the same number of planets:
    assert N_pl == np.sum(sss_per_sys['P_obs'] > 0) == np.sum(sss_per_sys['D_obs'] > 0) == np.sum(sss_per_sys['radii_obs'] > 0)
    assert N_pl == np.sum(sss_per_sys['tdur_obs'] >= 0) == np.sum(sss_per_sys['tdur_tcirc_obs'] >= 0)
    assert N_pl_pairs == np.sum(sss_per_sys['Rm_obs'] > 0) == np.sum(sss_per_sys['D_ratio_obs'] > 0) == np.sum(sss_per_sys['xi_obs'] >= 0) # NOTE: counting 'xi_obs' >= 0 can fail since values can have NaNs (when dividing by 0)?

    # Check that all fields in 'sss' have the right number of planets:
    keys = ['Rstar_obs', 'Mstar_obs', 'teff_obs', 'bp_rp_obs', 'e_bp_rp_obs', 'cdpp4p5_obs', 'P_obs', 'D_obs', 'tdur_obs', 'tdur_tcirc_obs', 'radii_obs']
    for key in keys:
        assert N_pl == len(sss[key])
    assert N_pl == len(sss['tdur_tcirc_1_obs']) + len(sss['tdur_tcirc_2p_obs'])
    assert N_pl == len(sss['D_above_obs']) + len(sss['D_below_obs'])

    assert N_pl_pairs == len(sss['Rm_obs'])
    assert N_pl_pairs == len(sss['D_ratio_obs']) == len(sss['D_ratio_above_obs']) + len(sss['D_ratio_below_obs']) + len(sss['D_ratio_across_obs'])
    assert N_pl_pairs == len(sss['xi_obs']) == len(sss['xi_res_obs']) + len(sss['xi_nonres_obs'])
    assert len(sss['xi_res_obs']) >= len(sss['xi_res32_obs']) + len(sss['xi_res21_obs'])

def test_compute_summary_stats_from_cat_obs(load_dir=loadfiles_directory, run_number=''):
    sss_per_sys, sss = compute_summary_stats_from_cat_obs(file_name_path=load_dir, run_number=run_number)
    run_test_summary_stats_cat_obs(sss_per_sys, sss)

def test_combine_sss_or_sssp_per_sys(load_dir=loadfiles_directory_multiple, rn1='1', rn2='2'):
    sss_per_sys1, sss1 = compute_summary_stats_from_cat_obs(file_name_path=load_dir, run_number=rn1)
    sss_per_sys2, sss2 = compute_summary_stats_from_cat_obs(file_name_path=load_dir, run_number=rn2)

    # Test each catalog first:
    run_test_summary_stats_cat_obs(sss_per_sys1, sss1)
    run_test_summary_stats_cat_obs(sss_per_sys2, sss2)

    # Test combined catalog:
    sss_per_sys_combined = combine_sss_or_sssp_per_sys(sss_per_sys1, sss_per_sys2)
    sss_combined = combine_sss_or_sssp_per_sys(sss1, sss2)
    run_test_summary_stats_cat_obs(sss_per_sys_combined, sss_combined)

def test_load_cat_phys_multiple_and_compute_combine_summary_stats(load_dir=loadfiles_directory_multiple, run_numbers=range(1,6)):
    assert len(run_numbers) > 1, 'Must load more than one catalog for this test.'
    N_sim_combined = 0
    for rn in run_numbers:
        N_sim, cos_factor, P_min, P_max, radii_min, radii_max = read_targets_period_radius_bounds(load_dir + 'periods_all%s.out' % rn)
        N_sim_combined += N_sim
    sssp_per_sys_combined, sssp_combined = load_cat_phys_multiple_and_compute_combine_summary_stats(load_dir, run_numbers=run_numbers, load_full_tables=True)

    # WARNING: assumes that each individual catalog has the same values 'P_min', 'P_max', etc.
    run_test_summary_stats_cat_phys(sssp_per_sys_combined, sssp_combined, N_sim=N_sim_combined, P_min=P_min, P_max=P_max, radii_min=radii_min, radii_max=radii_max)
